# type: ignore
"""

    Greynir: Natural language processing for Icelandic

    IFD tagger module

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module implements a class that handles conversion from Greynir's
    terminal variants to the Icelandic Frequency Dictionary (IFD) tagset.

    The tagset is described here: http://www.malfong.is/files/ot_tagset_files_is.pdf

"""

from typing import Any

from .settings import (
    UndeclinableAdjectives,
    StaticPhrases,
)
from .binparser import canonicalize_token
from .bintokenizer import TOK


class IFD_Tagset:

    """ Utility class to generate POS tags compatible with
        the Icelandic Frequency Dictionary (IFD) tagset
        (cf. http://www.malfong.is/files/ot_tagset_book_is.pdf) """

    # Strings that must be present in the grammatical form for variants
    BIN_TO_VARIANT = {
        "NF": "nf",  # Nefnifall / nominative
        "ÞF": "þf",  # Þolfall / accusative
        "ÞF2": "þf",  # Þolfall / accusative
        "ÞGF": "þgf",  # Þágufall / dative
        "ÞGF2": "þgf",  # Þágufall / dative
        "EF": "ef",  # Eignarfall / genitive
        "EF2": "ef",  # Eignarfall / genitive
        "KK": "kk",  # Karlkyn / masculine
        "KVK": "kvk",  # Kvenkyn / feminine
        "HK": "hk",  # Hvorugkyn / neutral
        "ET": "et",  # Eintala / singular
        "ET2": "et",  # Eintala / singular
        "ET3": "et",  # Eintala / singular
        "FT": "ft",  # Fleirtala / plural
        "FT2": "ft",  # Fleirtala / plural
        "FT3": "ft",  # Fleirtala / plural
        "FSB": "fsb",  # Frumstig, sterk beyging
        "FVB": "fvb",  # Frumstig, veik beyging
        "MST": "mst",  # Miðstig / comparative
        "MST2": "mst",  # Miðstig / comparative
        "ESB": "esb",  # Efsta stig, sterk beyging / superlative
        "EVB": "evb",  # Efsta stig, veik beyging / superlative
        "EST": "est",  # Efsta stig / superlative
        "EST2": "est",  # Efsta stig / superlative
        "1P": "p1",  # Fyrsta persóna / first person
        "2P": "p2",  # Önnur persóna / second person
        "3P": "p3",  # Þriðja persóna / third person
        "OP": "op",  # Ópersónuleg sögn
        "GM": "gm",  # Germynd
        "MM": "mm",  # Miðmynd
        "SB": "sb",  # Sterk beyging
        "VB": "vb",  # Veik beyging
        "NH": "nh",  # Nafnháttur
        "FH": "fh",  # Framsöguháttur
        "BH": "bh",  # Boðháttur
        "LH": "lh",  # Lýsingarháttur (nútíðar)
        "VH": "vh",  # Viðtengingarháttur
        "NT": "nt",  # Nútíð
        "ÞT": "þt",  # Þátíð
        "SAGNB": "sagnb",  # Sagnbót ('vera' -> 'hefur verið')
        "SAGNB2": "sagnb",
        "LHÞT": "lhþt",  # Lýsingarháttur þátíðar ('var lentur')
        "gr": "gr",  # Greinir
        "gr2": "gr",  # Greinir
    }

    # Create a list of BIN tags in descending order by length
    BIN_TAG_LIST = sorted(BIN_TO_VARIANT.keys(), key=lambda x: len(x), reverse=True)

    KIND_TO_TAG = {
        # !!! TBD: put in more precise tags
        "DATE": "to",
        "DATEREL": "to",
        "DATEABS": "to",
        "TIME": "to",
        "TIMESTAMP": "to",
        "TIMESTAMPREL": "to",
        "TIMESTAMPABS": "to",
        "PERCENT": "tp",
        "MEASUREMENT": "to",
    }

    CAT_TO_SCHEME = {
        "no": "_n",
        "kk": "_n",
        "kvk": "_n",
        "hk": "_n",
        "fn": "_f",
        "abfn": "_f",
        "pfn": "_f",
        "gr": "_g",
        "to": "_t",
        "töl": "_t",
        "tala": "_number",
        "raðnr": "_raðnr",
        "ártal": "_year",
        "so": "_s",
        "lo": "_l",
        "ao": "_a",
        "eo": "_a",
        "fs": "_a",
        "uh": "_a",
        "st": "_c",
        "stt": "_c",
        "nhm": "_c",
        "entity": "_e",
        "prósenta": "_t",
        "sérnafn": "_n",
        "fyrirtæki": "_n",
        "person": "_n",
        "gata": "_n",
    }

    FN_FL = {
        "sá": "a",
        "þessi": "a",
        "hinn": "a",
        "slíkur": "b",
        "sjálfur": "b",
        "samur": "b",
        "sami": "b",  # ætti að vera samur
        "þvílíkur": "b",
        "minn": "e",
        "þinn": "e",
        "sinn": "e",
        "vor": "e",
        "einhver": "o",
        "sérhver": "o",
        "nokkur": "o",
        "allnokkur": "o",
        "hvorugur": "o",
        "allur": "o",
        "mestallur": "o",
        "flestallur": "o",
        "sumur": "o",
        "enginn": "o",
        "margur": "o",
        "flestir": "o",  # ætti að vera margur
        "einn": "o",
        "annar": "o",
        "neinn": "o",
        "sitthvað": "o",
        "ýmis": "o",
        "fáeinir": "o",
        "báðir": "o",
        "hver": "s",
        "hvor": "s",
        "hvaða": "s",
        "hvílíkur": "s",
    }
    # Beygingarmyndir sem tilheyra bæði 'sá' og pfn.
    FN_SAMFALL = {
        "það",
        "því",
        "þess",
        "þau",
        "þeir",
        "þá",
        "þær",
        "þeim",
        "þeirra",
    }
    FN_BÆÐI = {"sá", "það"}
    FN_PK = {
        "ég": "1",
        "þú": "2",
        "hann": "k",
        "hún": "v",
        "hán": "h",
        "það": "h",
        "þér": "2",
        "vér": "1",
    }
    # Sjálfgefin fallstjórn forsetninga
    FS_FALL = {
        "af": "þ",
        "andspænis": "þ",
        "auk": "e",
        "austan": "e",
        "austur": "o",
        "að": "þ",
        "eftir": "o",  # þ kemur líka til greina
        "fjarri": "þ",
        "fram": "o",
        "framhjá": "þ",
        "frá": "þ",
        "fyrir": "o",  # þ kemur líka til greina
        "gagnvart": "þ",
        "gegn": "þ",
        "gegnt": "þ",
        "gegnum": "o",
        "handa": "þ",
        "handan": "e",
        "heim": "o",
        "hjá": "þ",
        "inn": "o",
        "innan": "o",  # e kemur líka til greina
        "innanundir": "þ",  # o kemur líka til greina (til skoðunar)
        "jafnfætis": "þ",
        "kringum": "o",
        "lengi": "e",
        "megin": "e",
        "með": "þ",  # o kemur líka til greina
        "meðal": "e",
        "meðfram": "þ",
        "milli": "e",
        "millum": "e",
        "mót": "þ",
        "móti": "þ",
        "neðan": "e",
        "niður": "o",
        "norðan": "e",
        "norður": "o",
        "nálægt": "þ",
        "nær": "þ",
        "ofan": "e",
        "sakir": "e",
        "samkvæmt": "þ",
        "samtímis": "þ",
        "snemma": "e",
        "sunnan": "e",
        "suður": "o",
        "síðla": "e",
        "sökum": "e",
        "til": "e",
        "um": "o",  # n kemur líka til greina (til skoðunar)
        "umfram": "o",
        "umhverfis": "o",
        "undan": "þ",
        "undir": "þ",  # o kemur líka til greina
        "upp": "o",
        "utan": "o",  # e kemur líka til greina
        "vegna": "e",
        "vestan": "e",
        "vestur": "o",
        "við": "o",  # þ kemur líka til greina
        "yfir": "o",  # þ kemur líka til greina
        "á": "o",  # þ kemur líka til greina
        "án": "e",
        "árla": "e",
        "ásamt": "þ",
        "í": "þ",  # o kemur líka til greina
        "öndvert": "þ",
        "úr": "þ",
        "út": "o",
    }
    # Raðtölur
    ORDINALS = frozenset(
        [
            "fyrstur",
            "annar",
            "þriðji",
            "fjórði",
            "fimmti",
            "sjötti",
            "sjöundi",
            "áttundi",
            "níundi",
            "tíundi",
            "ellefti",
            "tólfti",
            "þrettándi",
            "fjórtándi",
            "fimmtándi",
            "sextándi",
            "sautjándi",
            "átjándi",
            "nítjándi",
            "tuttugasti",
            "þrítugasti",
            "fertugasti",
            "fimmtugasti",
            "sextugasti",
            "sjötugasti",
            "átttugasti",
            "nítugasti",
            "hundraðasti",
            "tvöhundraðasti",
            "þrjúhundraðasti",
            "fjögurhundraðasti",
            "fimmhundraðasti",
            "sexhundraðasti",
            "sjöhundraðasti",
            "áttahundraðasti",
            "níuhundraðasti",
            "þúsundasti",
            "tvöþúsundasti",
            "þrjúþúsundasti",
            "fjögurþúsundasti",
            "fimmþúsundasti",
            "sexþúsundasti",
            "sjöþúsundasti",
            "áttaþúsundasti",
            "níuþúsundasti",
            "tíuþúsundasti",
            "milljónasti",
            "milljarðasti",
        ]
    )

    def __init__(self, *args: Any, **kwargs: Any):
        # Initialize the tagset from a token or from keyword parameters
        if len(args) == 1 and len(kwargs) == 0:
            # Single positional parameter: assume it's a token dict
            self._init_from(args[0])
        elif len(args) == 0:
            # Relay keyword parameters onwards
            self._init_from(kwargs)
        else:
            raise ValueError("Unsupported parameter list")

    def _init_from(self, t):
        """ Initialize the tagset parameters from a dict """
        self._cache = None
        self._kind = t.get("k")
        self._cat = t.get("c")
        self._fl = t.get("f")
        self._txt = t.get("x")
        if self._txt:
            self._txt = self._txt.lower()
        self._stem = t.get("s")
        self._v = t.get("v")
        if "t" in t:
            # Terminal: assemble the variants
            a = t["t"].split("_")
            self._first = a[0]
            self._tagset = set(a[1:])
        else:
            self._first = None
            self._tagset = set()
        if self._cat in {"kk", "kvk", "hk"}:
            self._tagset.add(self._cat)
        if "b" in t:
            # Mix the BIN tags into the set
            beyging = t["b"]
            for bin_tag in self.BIN_TAG_LIST:
                # This loop proceeds in descending order by tag length
                if bin_tag in beyging:
                    self._tagset.add(self.BIN_TO_VARIANT[bin_tag])
                    beyging = beyging.replace(bin_tag, "").replace("--", "")
                    if not beyging:
                        break

    def _tagstring(self):
        """ Calculate the IFD tagstring from the tagset """
        if self._kind == "PUNCTUATION":
            return self._txt
        if self._kind in self.KIND_TO_TAG:
            return self.KIND_TO_TAG[self._kind]
        key = self._first or self._cat or self._kind
        scheme = self.CAT_TO_SCHEME.get(key)
        if scheme is None:
            return "[" + key + "]"  # !!! TODO
        func = getattr(self, scheme)
        return scheme[1:] if func is None else func()

    def has_tag(self, tag):
        """ Returns True if the tagset contains the given feature,
            i.e. kk, kvk, þgf, nh, p3, vb, etc. """
        return tag in self._tagset

    def __str__(self):
        """ Return the tags formatted as an IFD compatible string """
        if self._cache is None:
            self._cache = self._tagstring()
        return self._cache

    def _n(self):
        return (
            "n"
            + self._kyn()
            + self._tala()
            + self._fall()
            + self._greinir()
            + self._sérnöfn()
        )

    def _l(self):
        return (
            "l"
            + self._kyn()
            + self._tala()
            + self._fall()
            + self._beyging()
            + self._stig()
        )

    def _f(self):
        return (
            "f" + self._flokkur_f() + self._kyn_persóna() + self._tala() + self._fall()
        )

    def _g(self):
        return "g" + self._kyn() + self._tala() + self._fall()

    def _t(self):
        return (
            "t"
            + self._flokkur_t()
            + self._kyn()
            + self._tala(default="f")
            + self._fall()
        )

    def _s(self):
        if "lh" in self._tagset and "nt" in self._tagset:
            # Lýsingarháttur nútíðar
            return "slg"  # Alltaf germynd - gæti hugsanlega verið miðmynd
        if "lhþt" in self._tagset:
            # Lýsingarháttur þátíðar
            return "sþ" + self._mynd() + self._kyn() + self._tala() + self._fall()
        if "nh" in self._tagset:
            # Nafnháttur
            if "þt" in self._tagset:
                return "sn" + self._mynd() + "--þ"
            return "sn" + self._mynd()
        if "bh" in self._tagset:
            # Boðháttur
            return "sb" + self._mynd() + "2" + self._tala() + "n"  # Alltaf 2.p. nútíð
        if "sagnb" in self._tagset:
            # Sagnbót
            return "ss" + self._mynd()
        return (
            "s"
            + self._háttur()
            + self._mynd()
            + self._persóna()
            + self._tala()
            + self._tíð()
        )

    def _a(self):
        return "a" + self._flokkur_a() + self._stig_a()

    def _c(self):
        return "c" + self._flokkur_c()

    def _e(self):
        if self._txt[0].isupper():
            return "nxex-s"  # Sérnafn, óþekkt kyn, óþekkt fall
        return "e"

    def _x(self):
        return "x"

    def _number(self):
        return "tfkfn" if self._v == 11 or self._v % 10 != 1 else "tfken"

    def _raðnr(self):
        if self._tagset:
            return "l" + self._kyn() + "e" + self._fall() + "vf"
        # Lýsingarorð, eintala, veik beyging, frumstig. Kyn og fall óþekkt
        return "lxexvf"

    def _year(self):
        return "ta"

    def _kyn(self):
        if "kk" in self._tagset:
            return "k"
        if "kvk" in self._tagset:
            return "v"
        if "hk" in self._tagset:
            return "h"
        return "x"

    def _tala(self, default="e"):
        if "ft" in self._tagset:
            return "f"
        elif "et" in self._tagset:
            return "e"
        return default

    def _fall(self, default="n"):
        if "nf" in self._tagset:
            return "n"
        if "þf" in self._tagset:
            return "o"
        if "þgf" in self._tagset:
            return "þ"
        if "ef" in self._tagset:
            return "e"
        return default

    def _greinir(self):
        return "g" if "gr" in self._tagset else ""

    def _sérnöfn(self):
        if not self._stem:
            return ""
        if self._fl in {"örn", "göt", "heö", "bær", "þor", "lönd"}:
            return "ö" if "gr" in self._tagset else "-ö"
        if self._kind == "PERSON":
            return "-m"
        if self._kind == "CURRENCY":
            # !!! TBD
            return "e" if "gr" in self._tagset else "-e"
        if self._stem[0].isupper():
            # Sérnafn
            return "s" if "gr" in self._tagset else "-s"
        return ""

    def _stig(self):
        if "esb" in self._tagset or "evb" in self._tagset:
            return "e"
        if "mst" in self._tagset:
            return "m"
        return "f"

    def _beyging(self):
        if self._stem in UndeclinableAdjectives.ADJECTIVES:
            return "o"
        if "fsb" in self._tagset or "esb" in self._tagset:
            return "s"
        if "fvb" in self._tagset or "evb" in self._tagset or "mst" in self._tagset:
            return "v"
        if self._stem in self.ORDINALS:
            return "v"
        return "o"

    def _flokkur_f(self):
        if self._cat == "abfn":
            return "p"  # OTB flokkar abfn. með pfn.
        if self._cat == "pfn":
            return "p"
        if self._txt in self.FN_SAMFALL and self._stem in self.FN_BÆÐI:
            return "p"
        return self.FN_FL.get(self._stem, "x")

    def _kyn_persóna(self):
        if self._stem in self.FN_PK:
            return self.FN_PK[self._stem]
        if "kk" in self._tagset:
            return "k"
        if "kvk" in self._tagset:
            return "v"
        if "hk" in self._tagset:
            return "h"
        if "p1" in self._tagset:
            return "1"
        if "p2" in self._tagset:
            return "2"
        return "x"

    def _flokkur_t(self):
        if self._kind == "PERCENT":
            return "p"
        return "f"

    def _mynd(self):
        return "m" if "mm" in self._tagset else "g"

    def _háttur(self):
        # if "nh" in self._tagset:
        #    return "n"
        if "bh" in self._tagset:
            return "b"
        if "vh" in self._tagset:
            return "v"
        # if "sagnb" in self._tagset:
        #    return "s"
        if "lh" in self._tagset and "nt" in self._tagset:
            return "l"
        return "f"

    def _tíð(self):
        return "þ" if "þt" in self._tagset else "n"

    def _persóna(self):
        if "op" in self._tagset:
            return "3"
        if "p1" in self._tagset:
            return "1"
        if "p2" in self._tagset:
            return "2"
        return "3"

    def _stig_a(self):
        if "mst" in self._tagset:
            return "m"
        if "est" in self._tagset:
            return "e"
        return ""

    def _flokkur_a(self):
        if self._cat == "uh":
            return "u"
        if self._cat == "fs":
            return self._fall(default=self.FS_FALL.get(self._stem, "o"))
        return "a"

    def _flokkur_c(self):
        if self._first == "stt":
            # 'sem', 'er' as connective conjunctions
            return "t"
        if self._cat == "nhm":
            return "n"
        return ""

    @classmethod
    def word_tag_stream(cls, sentence_stream):
        """ Generator of a (word, tag) stream from a raw token stream.
            Both the input and the output streams are segmented into
            sentences. """
        for sent in sentence_stream:
            if not sent:
                continue
            output = []
            for t in sent:
                x = t.get("x")
                if not x:
                    continue
                if t.get("k", TOK.WORD) == TOK.PUNCTUATION:
                    output.append((x, x))
                    continue
                canonicalize_token(t)
                if " " in x:
                    lower_x = x.lower()
                    if StaticPhrases.has_details(lower_x):
                        # This is a static multi-word phrase
                        tags = StaticPhrases.tags(lower_x)
                        assert tags is not None
                        output.extend(zip(x.split(), tags))
                    else:
                        tag = str(cls(t))
                        for part in x.split():
                            # !!! TODO: this needs to be made more intelligent and detailed
                            if part in {"og", "eða"}:
                                output.append((part, "c"))
                            else:
                                output.append((part, tag or "Unk"))
                else:
                    tag = str(cls(t))
                    output.append((x, tag or "Unk"))
            if output:
                yield output



# type: ignore
"""

    test_original.py

    Tests for Greynir module

    Copyright (C) 2021 Miðeind ehf.
    Original author: Vilhjálmur Þorsteinsson

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

import pytest  # type: ignore

from reynir import Greynir
from reynir.bintokenizer import tokenize


@pytest.fixture(scope="module")
def r():
    """ Provide a module-scoped Greynir instance as a test fixture """
    r = Greynir()
    yield r
    # Do teardown here
    r.__class__.cleanup()


def test_original(r: Greynir) -> None:

    s = "Ég keypti     1000     EUR þann   23.   5. 2011 og græddi   10,5   %  ."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "   Friðjón    Pálsson hitti Friðbert   \tJ.   Ástráðsson í gær."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = " \t Casey   Holdman  \n  og   Luke    Skywalker   fóru saman   á bar  ."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Hver á    USD   5,75  sem ég fann í  grasinu með   5,558   prósent?"
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Virkjunin var  \t 600    MW  og var á Reynimel  40C  í Reykjavík  ."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Katrín   Júlíusdóttir   var   iðnaðar-  \n\t  og \t\t  viðskiptaráðherra"
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Friðbert  Marsillíus   Jónsson  keypti hlutabréf  í   Eimskip   hf.  fyrir    100    milljónir   í gær"
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Jens \tStoltenberg   keypti hlutabréf  nú   síðdegis  fyrir   100   milljónir  króna  kl. 12:30 30.   júlí 2002  og Jens   er stoltur   af því."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "Gengi  danskrar  krónu  féll um   2.000    EUR  kl.   14:00   30.  desember  ."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "Dómsmála-  ,    iðnaðar-   og  viðskiptaráðherra gerði víðreist um landið"
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Dagur   Bergþóruson   Eggertsson    hefur verið  farsæll borgarstjóri ."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Formaður framkvæmdastjórnarinnar er Ursula  \t\t   van  der    Leyen  ."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)

    s = "  Angela   Merkel   hefur   lengi   vel  verið kanslari   V-Þýskalands ."
    tlist = list(tokenize(s))
    assert sum(len(t.original or "") for t in tlist) == len(s)



if __name__ == "__main__":
    # When invoked as a main module, do a verbose test
    from reynir import Greynir
    greynir = Greynir()
    test_original(greynir)
    greynir.__class__.cleanup()



# type: ignore
"""

    test_cases.py

    Tests for Greynir module

    Copyright (C) 2021 Miðeind ehf.
    Original author: Vilhjálmur Þorsteinsson

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

import pytest  # type: ignore

from reynir import Greynir


@pytest.fixture(scope="module")
def r():
    """ Provide a module-scoped Greynir instance as a test fixture """
    r = Greynir()
    yield r
    # Do teardown here
    r.__class__.cleanup()


def test_cases(r: Greynir) -> None:
    s = r.parse_single("Ég átti svakalega stóran hest með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "svakalega stór hestur með fallegasta makkann"
    assert np_obj.accusative_np == 'svakalega stóran hest með fallegasta makkann'
    assert np_obj.dative_np == 'svakalega stórum hesti með fallegasta makkann'
    assert np_obj.genitive_np == 'svakalega stórs hests með fallegasta makkann'

    s = r.parse_single("Ég átti svakalega stóra hestinn með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "svakalega stóri hesturinn með fallegasta makkann"
    assert np_obj.accusative_np == 'svakalega stóra hestinn með fallegasta makkann'
    assert np_obj.dative_np == 'svakalega stóra hestinum með fallegasta makkann'
    assert np_obj.genitive_np == 'svakalega stóra hestsins með fallegasta makkann'

    s = r.parse_single("Ég átti hinn svakalega stóra hest með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "hinn svakalega stóri hestur með fallegasta makkann"
    assert np_obj.accusative_np == 'hinn svakalega stóra hest með fallegasta makkann'
    assert np_obj.dative_np == 'hinum svakalega stóra hesti með fallegasta makkann'
    assert np_obj.genitive_np == 'hins svakalega stóra hests með fallegasta makkann'

    s = r.parse_single("Ég átti svakalega stóra hesta með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "svakalega stórir hestar með fallegasta makkann"
    assert np_obj.accusative_np == 'svakalega stóra hesta með fallegasta makkann'
    assert np_obj.dative_np == 'svakalega stórum hestum með fallegasta makkann'
    assert np_obj.genitive_np == 'svakalega stórra hesta með fallegasta makkann'

    s = r.parse_single("Ég átti svakalega stóru hestana með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "svakalega stóru hestarnir með fallegasta makkann"
    assert np_obj.accusative_np == 'svakalega stóru hestana með fallegasta makkann'
    assert np_obj.dative_np == 'svakalega stóru hestunum með fallegasta makkann'
    assert np_obj.genitive_np == 'svakalega stóru hestanna með fallegasta makkann'

    s = r.parse_single("Ég átti hina svakalega stóru hesta með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "hinir svakalega stóru hestar með fallegasta makkann"
    assert np_obj.accusative_np == 'hina svakalega stóru hesta með fallegasta makkann'
    assert np_obj.dative_np == 'hinum svakalega stóru hestum með fallegasta makkann'
    assert np_obj.genitive_np == 'hinna svakalega stóru hesta með fallegasta makkann'

    s = r.parse_single("Ég átti svakalega stóran hest með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "svakalega stór hestur með fallegasta makkann"
    assert np_obj.accusative_np == 'svakalega stóran hest með fallegasta makkann'
    assert np_obj.dative_np == 'svakalega stórum hesti með fallegasta makkann'
    assert np_obj.genitive_np == 'svakalega stórs hests með fallegasta makkann'

    s = r.parse_single("Ég átti allra stærsta hestinn með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "allra stærsti hesturinn með fallegasta makkann"
    assert np_obj.accusative_np == 'allra stærsta hestinn með fallegasta makkann'
    assert np_obj.dative_np == 'allra stærsta hestinum með fallegasta makkann'
    assert np_obj.genitive_np == 'allra stærsta hestsins með fallegasta makkann'

    s = r.parse_single("Ég átti hinn allra stærsta hest með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "hinn allra stærsti hestur með fallegasta makkann"
    assert np_obj.accusative_np == 'hinn allra stærsta hest með fallegasta makkann'
    assert np_obj.dative_np == 'hinum allra stærsta hesti með fallegasta makkann'
    assert np_obj.genitive_np == 'hins allra stærsta hests með fallegasta makkann'

    s = r.parse_single("Ég átti allra stærsta hesta með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "allra stærstir hestar með fallegasta makkann"
    assert np_obj.accusative_np == 'allra stærsta hesta með fallegasta makkann'
    assert np_obj.dative_np == 'allra stærstum hestum með fallegasta makkann'
    assert np_obj.genitive_np == 'allra stærstra hesta með fallegasta makkann'

    s = r.parse_single("Ég átti allra stærstu hestana með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "allra stærstu hestarnir með fallegasta makkann"
    assert np_obj.accusative_np == 'allra stærstu hestana með fallegasta makkann'
    assert np_obj.dative_np == 'allra stærstu hestunum með fallegasta makkann'
    assert np_obj.genitive_np == 'allra stærstu hestanna með fallegasta makkann'

    s = r.parse_single("Ég átti hina allra stærstu hesta með fallegasta makkann.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "hinir allra stærstu hestar með fallegasta makkann"
    assert np_obj.accusative_np == 'hina allra stærstu hesta með fallegasta makkann'
    assert np_obj.dative_np == 'hinum allra stærstu hestum með fallegasta makkann'
    assert np_obj.genitive_np == 'hinna allra stærstu hesta með fallegasta makkann'

    s = r.parse_single("Ég átti allra stærsta hestinn sem kunni fimm gangtegundir.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "allra stærsti hesturinn sem kunni fimm gangtegundir"
    assert np_obj.accusative_np == 'allra stærsta hestinn sem kunni fimm gangtegundir'
    assert np_obj.dative_np == 'allra stærsta hestinum sem kunni fimm gangtegundir'
    assert np_obj.genitive_np == 'allra stærsta hestsins sem kunni fimm gangtegundir'

    s = r.parse_single("Ég átti hinn allra stærsta hest sem kunni fimm gangtegundir.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "hinn allra stærsti hestur sem kunni fimm gangtegundir"
    assert np_obj.accusative_np == 'hinn allra stærsta hest sem kunni fimm gangtegundir'
    assert np_obj.dative_np == 'hinum allra stærsta hesti sem kunni fimm gangtegundir'
    assert np_obj.genitive_np == 'hins allra stærsta hests sem kunni fimm gangtegundir'

    s = r.parse_single("Ég átti allra stærsta hesta sem kunnu fimm gangtegundir.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "allra stærstir hestar sem kunnu fimm gangtegundir"
    assert np_obj.accusative_np == 'allra stærsta hesta sem kunnu fimm gangtegundir'
    assert np_obj.dative_np == 'allra stærstum hestum sem kunnu fimm gangtegundir'
    assert np_obj.genitive_np == 'allra stærstra hesta sem kunnu fimm gangtegundir'

    s = r.parse_single("Ég átti allra stærstu hestana sem kunnu fimm gangtegundir.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "allra stærstu hestarnir sem kunnu fimm gangtegundir"
    assert np_obj.accusative_np == 'allra stærstu hestana sem kunnu fimm gangtegundir'
    assert np_obj.dative_np == 'allra stærstu hestunum sem kunnu fimm gangtegundir'
    assert np_obj.genitive_np == 'allra stærstu hestanna sem kunnu fimm gangtegundir'

    s = r.parse_single("Ég átti hina allra stærstu hesta sem kunnu fimm gangtegundir.")
    np_obj = s.tree.S_MAIN.IP.VP.NP_OBJ
    assert np_obj.nominative_np == "hinir allra stærstu hestar sem kunnu fimm gangtegundir"
    assert np_obj.accusative_np == 'hina allra stærstu hesta sem kunnu fimm gangtegundir'
    assert np_obj.dative_np == 'hinum allra stærstu hestum sem kunnu fimm gangtegundir'
    assert np_obj.genitive_np == 'hinna allra stærstu hesta sem kunnu fimm gangtegundir'

    s = r.parse_single("Pál, hinn vinsæla landsliðsmann sem spilaði þrjátíu leiki "
        "með landsliðinu á sínum tíma, langar að leggja skóna á hilluna.")
    assert (s.tree.S_MAIN.IP.NP_SUBJ.nominative_np ==
        'Páll , hinn vinsæli landsliðsmaður sem spilaði þrjátíu leiki með landsliðinu á sínum tíma'
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.accusative_np ==
        'Pál , hinn vinsæla landsliðsmann sem spilaði þrjátíu leiki með landsliðinu á sínum tíma'
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.dative_np ==
        'Páli , hinum vinsæla landsliðsmanni sem spilaði þrjátíu leiki með landsliðinu á sínum tíma'
    )
    assert(s.tree.S_MAIN.IP.NP_SUBJ.genitive_np ==
        'Páls , hins vinsæla landsliðsmanns sem spilaði þrjátíu leiki með landsliðinu á sínum tíma'
    )

    s = r.parse_single("Pósturinn Páll, hinn sívinsæli gleðigjafi, er á dagskrá í sumar.")
    assert (s.tree.S_MAIN.IP.NP_SUBJ.nominative_np ==
        "Pósturinn Páll , hinn sívinsæli gleðigjafi"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.accusative_np ==
        "Póstinn Pál , hinn sívinsæla gleðigjafa"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.dative_np ==
        "Póstinum Páli , hinum sívinsæla gleðigjafa"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.genitive_np ==
        "Póstsins Páls , hins sívinsæla gleðigjafa"
    )

    s = r.parse_single("Pósturinn Páll og kötturinn Njáll, tveir sívinsælir gleðigjafar, eru á dagskrá í sumar.")
    assert (s.tree.S_MAIN.IP.NP_SUBJ.nominative_np ==
        "Pósturinn Páll og kötturinn Njáll , tveir sívinsælir gleðigjafar"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.accusative_np ==
        "Póstinn Pál og köttinn Njál , tvo sívinsæla gleðigjafa"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.dative_np ==
        "Póstinum Páli og kettinum Njáli , tveimur sívinsælum gleðigjöfum"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.genitive_np ==
        "Póstsins Páls og kattarins Njáls , tveggja sívinsælla gleðigjafa"
    )

    s = r.parse_single("Rauð viðvörun hefur verið gefin út.")
    assert (s.tree.S_MAIN.IP.NP_SUBJ.nominative_np ==
        "Rauð viðvörun"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.accusative_np ==
        "Rauða viðvörun"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.dative_np ==
        "Rauðri viðvörun"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.genitive_np ==
        "Rauðrar viðvörunar"
    )

    s = r.parse_single("Rauða viðvörunin hefur verið gefin út.")
    assert (s.tree.S_MAIN.IP.NP_SUBJ.nominative_np ==
        "Rauða viðvörunin"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.accusative_np ==
        "Rauðu viðvörunina"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.dative_np ==
        "Rauðu viðvöruninni"
    )
    assert (s.tree.S_MAIN.IP.NP_SUBJ.genitive_np ==
        "Rauðu viðvörunarinnar"
    )


def test_noun_phrases(r):
    """ Test functions for easy manipulation of noun phrases """
    
    # Doesn't work for some reason
    np = r.parse_noun_phrase("þrír lúxus-miðar á Star Wars í dag")
    assert np.tree is not None
    assert np.nominative == "þrír lúxus-miðar á Star Wars í dag"
    assert np.accusative == "þrjá lúxus-miða á Star Wars í dag"
    assert np.dative == "þremur lúxus-miðum á Star Wars í dag"
    assert np.genitive == "þriggja lúxus-miða á Star Wars í dag"
    assert np.indefinite == "þrír lúxus-miðar á Star Wars í dag"
    assert np.canonical == "lúxus-miði"

    from reynir import NounPhrase
    np = NounPhrase(
        "þrír glæsilegir lúxus-bíómiðar á Star Wars "
        "og að auki tveir stútfullir pokar af ilmandi poppi"
    )
    assert np.parsed
    assert len(np) == len(str(np))
    assert (
        "Hér er kvittunin þín fyrir {np:þgf}. "
        "Þar með ertu búin(n) að kaupa {np:þf}.".format(np=np)
        == "Hér er kvittunin þín fyrir þremur glæsilegum lúxus-bíómiðum "
        "á Star Wars og að auki tveimur stútfullum pokum af ilmandi poppi. "
        "Þar með ertu búin(n) að kaupa þrjá glæsilega lúxus-bíómiða "
        "á Star Wars og að auki tvo stútfulla poka af ilmandi poppi."
    )
    np = NounPhrase('skjótti hesturinn')
    assert np.parsed
    assert np.case == "nf"
    assert np.person == "p3"
    assert np.number == "et"
    assert np.gender == "kk"
    assert str(np) == "skjótti hesturinn"
    assert "Hér er {np:nf}".format(np=np) == 'Hér er skjótti hesturinn'
    assert "Um {np:þf}".format(np=np) == 'Um skjótta hestinn'
    assert "Frá {np:þgf}".format(np=np) == 'Frá skjótta hestinum'
    assert "Til {np:ef}".format(np=np) == 'Til skjótta hestsins'
    assert "Hér er {np:ángr}".format(np=np) == 'Hér er skjóttur hestur'
    np = NounPhrase("þrír skjóttir hestar")
    assert np.parsed
    assert np.number == "ft"
    assert np.case == "nf"
    assert np.person == "p3"
    assert np.gender == "kk"
    assert str(np) == "þrír skjóttir hestar"
    assert len(np) == len(str(np))
    assert "Umræðuefnið er {np:stofn}".format(np=np) == 'Umræðuefnið er skjóttur hestur'
    try:
        "Óleyfilegt {np:.2f}".format(np=np)
    except ValueError:
        pass
    else:
        assert False, "Should have raised ValueError"
    try:
        "Óleyfilegt {np:abc}".format(np=np)
    except ValueError:
        pass
    else:
        assert False, "Should have raised ValueError"
    np = NounPhrase("Doddi át kökuna")
    assert not np.parsed
    assert np.gender is None
    assert np.number is None
    assert np.case is None
    assert np.person is None
    np = NounPhrase("")
    assert not np.parsed
    assert np.gender is None
    assert np.number is None
    assert np.case is None
    assert np.person is None
    np = NounPhrase("buxur")
    assert np.parsed
    assert np.number == "ft"
    assert np.genitive == "buxna"
    np = NounPhrase("buxur", force_number="et")
    assert not np.parsed
    np = NounPhrase("gleraugun hans afa", force_number="ft")
    assert np.parsed
    assert np.number == "ft"
    assert np.genitive == "gleraugnanna hans afa"
    np = NounPhrase("gleraugun hans afa", force_number="et")
    assert not np.parsed
    np = NounPhrase("eyjar", force_number="ft")
    assert np.parsed
    assert np.number == "ft"
    assert np.nominative == "eyjar"
    np = NounPhrase("eyjar", force_number="et")
    assert np.parsed
    assert np.number == "et"
    assert np.nominative == "ey"
    np = NounPhrase("eyjar", force_number="plural")
    assert np.parsed
    assert np.number == "ft"
    assert np.nominative == "eyjar"
    np = NounPhrase("eyjar", force_number="singular")
    assert np.parsed
    assert np.number == "et"
    assert np.nominative == "ey"
    np = NounPhrase("Spánn")
    assert np.parsed
    assert np.nominative == "Spánn"
    assert np.genitive == "Spánar"


def test_addresses():
    from reynir import NounPhrase
    np = NounPhrase("Laugavegi 20b")
    assert np.nominative == "Laugavegur 20b"
    assert np.accusative == "Laugaveg 20b"
    assert np.dative == "Laugavegi 20b"
    assert np.genitive == "Laugavegar 20b"
    np = NounPhrase("Baugatanga 6C")
    assert np.nominative == "Baugatangi 6C"
    assert np.accusative == "Baugatanga 6C"
    assert np.dative == "Baugatanga 6C"
    assert np.genitive == "Baugatanga 6C"
    np = NounPhrase("Æðarvarp 17")
    assert np.nominative == "Æðarvarp 17"
    assert np.accusative == "Æðarvarp 17"
    assert np.dative == "Æðarvarpi 17"
    assert np.genitive == "Æðarvarps 17"
    np = NounPhrase("Vættaborgir 17")
    assert np.nominative == "Vættaborgir 17"
    assert np.accusative == "Vættaborgir 17"
    assert np.dative == "Vættaborgum 17"
    assert np.genitive == "Vættaborga 17"
    np = NounPhrase("Vættaborga 17")
    assert np.nominative == "Vættaborgir 17"
    assert np.accusative == "Vættaborgir 17"
    assert np.dative == "Vættaborgum 17"
    assert np.genitive == "Vættaborga 17"
    np = NounPhrase("Sólarsölum 3B")
    assert np.nominative == "Sólarsalir 3B"
    assert np.accusative == "Sólarsali 3B"
    assert np.dative == "Sólarsölum 3B"
    assert np.genitive == "Sólarsala 3B"
    np = NounPhrase("Gamla vitanum í Garðskaga á Reykjanesi")
    assert np.nominative == "Gamli vitinn í Garðskaga á Reykjanesi"
    assert np.accusative == "Gamla vitann í Garðskaga á Reykjanesi"
    assert np.dative == "Gamla vitanum í Garðskaga á Reykjanesi"
    assert np.genitive == "Gamla vitans í Garðskaga á Reykjanesi"
    np = NounPhrase("Gamla vitans í Garðskaga á Reykjanesi")
    assert np.nominative == "Gamli vitinn í Garðskaga á Reykjanesi"
    assert np.accusative == "Gamla vitann í Garðskaga á Reykjanesi"
    assert np.dative == "Gamla vitanum í Garðskaga á Reykjanesi"
    assert np.genitive == "Gamla vitans í Garðskaga á Reykjanesi"


if __name__ == "__main__":
    # When invoked as a main module, do a verbose test
    from reynir import Greynir
    greynir = Greynir()
    test_cases(greynir)
    test_noun_phrases(greynir)
    test_addresses()
    greynir.__class__.cleanup()



"""

    Greynir: Natural language processing for Icelandic

    High-level wrapper for the Greynir tokenizer, parser and reducer

    Copyright (C) 2021 Miðeind ehf.
    Original author: Vilhjálmur Þorsteinsson

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module implements a high-level interface to the Greynir
    tokenizer, parser and reducer for parsing Icelandic text into
    trees.

"""

from typing import (
    Any,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Union,
    Callable,
    Dict,
    List,
    Tuple,
    NamedTuple,
    Type,
    cast,
)
from typing_extensions import TypedDict

import time
import operator
import json
from threading import Lock

from tokenizer import Tok, correct_spaces, paragraphs, mark_paragraphs

from .bintokenizer import (
    StringIterable,
    TokenList,
    CanonicalTokenDict,
    tokenize as bin_tokenize,
    tokens_are_foreign,
    load_token,
)
from .binparser import BIN_Token
from .fastparser import Fast_Parser, ParseError
from .reducer import Reducer
from .cache import cached_property
from .simpletree import SimpleTree
from .incparser import ICELANDIC_RATIO
from .lemmatize import LemmaTuple, Comparable, simple_lemmatize


# The type of the values generated by the tokenizer.paragraphs() function
SentenceTuple = Tuple[int, TokenList]
GetterFunc = Callable[[SimpleTree], str]
GreynirType = Type["Greynir"]

# The Sentence.terminals attribute returns a list of Terminal objects
Terminal = NamedTuple(
    "Terminal",
    [
        ("text", str),
        ("lemma", str),
        ("category", str),
        ("variants", List[str]),
        ("index", int),
    ],
)

# Progress function parameter type
ProgressFunc = Optional[Callable[[float], None]]

# The type of a parse result
class ParseResult(TypedDict):
    sentences: List["_Sentence"]
    num_sentences: int
    num_parsed: int
    num_tokens: int
    ambiguity: float
    parse_time: float
    reduce_time: float

# The default maximum length of a sentence, in tokens, that we attempt to parse
DEFAULT_MAX_SENT_TOKENS = 90


class _Sentence:

    """ A container for a sentence that has been extracted from the
        tokenizer. The sentence can be explicitly parsed by calling
        sentence.parse(). After parsing, a number of query functions
        are available on the parse tree. """

    def __init__(self, job: "_Job", s: TokenList) -> None:
        """ NOTE! If attributes are added here, the _Sentence.load() function
            below needs to be updated accordingly. """
        self._job = job
        # s is a token list
        self._s = s
        self._len = len(s)
        assert self._len > 0  # Input should be already sanitized
        self._err_index: Optional[int] = None
        self._error: Optional[ParseError] = None
        self._simplified_tree: Optional[SimpleTree] = None
        self._tree = None
        # Number of possible combinations
        self._num: Optional[int] = None
        # Score of best parse tree
        self._score: Optional[int] = None
        # Cached terminals
        self._terminals: Optional[List[Terminal]] = None
        if self._job.parse_immediately:
            # We want an immediate parse of the sentence
            self.parse()

    def __len__(self) -> int:
        """ Return the number of tokens in the sentence """
        return self._len

    def parse(self) -> bool:
        """ Parse the sentence """
        if self._num is not None:
            # Already parsed
            return self._num > 0
        job = self._job
        num = 0
        score = 0
        tree = None
        try:
            # Invoke the parser on the sentence tokens
            tree, num, score = job.parse(self._s)
        except ParseError as e:
            self._err_index = self._len - 1 if e.token_index is None else e.token_index
            self._error = e
        self._tree = tree
        if tree is None:
            self._simplified_tree = None
        else:
            # Create a simplified tree as well
            self._simplified_tree = SimpleTree.from_deep_tree(tree, self._s)
        self._num = num
        self._score = score
        return num > 0

    @property
    def error(self) -> Optional[ParseError]:
        """ Return the ParseError that occurred when parsing this sentence, or None """
        return self._error

    @property
    def err_index(self) -> Optional[int]:
        """ Return the index of the error token, if an error occurred;
            otherwise None """
        return self._err_index

    @property
    def tokens(self) -> TokenList:
        """ Return the tokens in the sentence """
        return self._s

    def is_foreign(self, min_icelandic_ratio: float = ICELANDIC_RATIO) -> bool:
        """ Return True if the sentence is probably not in Icelandic """
        return tokens_are_foreign(self.tokens, min_icelandic_ratio)

    @property
    def combinations(self) -> Optional[int]:
        """ Return the number of different parse tree combinations
            for the sentence, or 0 if no parse tree was found,
            or None if the sentence hasn't been parsed """
        return self._num

    @property
    def score(self) -> Optional[int]:
        """ The score of the best parse tree for the sentence """
        return self._score

    @property
    def tree(self) -> Optional[SimpleTree]:
        """ Return the simplified parse tree, or None
            if the sentence hasn't been parsed """
        return self._simplified_tree

    @property
    def deep_tree(self) -> Any:
        """ Return the original deep tree, as constructed by the parser,
            corresponding directly to grammar nonterminals and terminals """
        return self._tree

    @property
    def flat_tree(self) -> Optional[str]:
        """ Return a flat text representation of the simplified parse tree """
        return None if self.tree is None else self.tree.flat

    @cached_property
    def text(self) -> str:
        """ Return a raw text representation of the sentence,
            with spaces between all tokens """
        return " ".join(t.txt for t in self._s if t.txt)

    @property
    def tidy_text(self) -> str:
        """ Return a [more] correctly spaced text representation of the sentence """
        if self.tree is None:
            txt = self.text
        else:
            terminals = self.terminal_nodes
            # Use the terminal text representation -
            # it's got fancy em/en-dashes and stuff
            txt = " ".join(t.text for t in terminals)
        return correct_spaces(txt)

    @property
    def terminals(self) -> Optional[List[Terminal]]:
        """ Return a list of tuples, one for each terminal in the sentence.
            The tuples contain the original text of the token that matched
            the terminal, the associated word lemma, the category, and a set
            of variants (case, number, gender, etc.) """
        if self.tree is None:
            # Must parse the sentence first, without errors
            return None
        if self._terminals is not None:
            # Already calculated and cached
            return self._terminals
        # Generate the terminal list from the parse tree
        # pylint: disable=not-an-iterable
        self._terminals = [
            Terminal(d.text, d.lemma, d.tcat, d.all_variants, d.index or 0)
            for d in self.terminal_nodes
        ]
        return self._terminals

    @cached_property
    def terminal_nodes(self) -> List[SimpleTree]:
        """ Return a list of the terminal nodes within the parse tree
            for this sentence """
        if self.tree is None:
            return []
        return [d for d in self.tree.descendants if d.is_terminal]

    @property
    def lemmas(self) -> Optional[List[str]]:
        """ Convenience property to return the lemmas only """
        t = self.terminals
        return None if t is None else [terminal.lemma for terminal in t]

    @property
    def lemmas_mm(self) -> List[str]:
        """ Convenience property to return the lemmas only, in middle voice
            if the terminal so specifies """
        t = self.terminals
        result: List[str] = []
        if t:
            for terminal in t:
                if terminal.category != "so" or "mm" not in terminal.variants:
                    # Not a middle voice verb: return the normal lemma
                    result.append(terminal.lemma)
                else:
                    # Middle voice verb: return the middle voice lemma
                    # (ending in -st, such as 'dást' for the verb 'dá')
                    result.append(BIN_Token.mm_verb_stem(terminal.lemma))
        return result

    @property
    def categories(self) -> Optional[List[str]]:
        """ Convenience property to return the categories only """
        if self.tree is None:
            return None
        # Note that here we return the BÍN category,
        # not the terminal category (tcat)
        # pylint: disable=not-an-iterable
        return [d.cat for d in self.terminal_nodes]

    @property
    def lemmas_and_cats(self) -> Optional[List[Tuple[str, str]]]:
        """ Convenience property to return (lemma, category) tuples """
        if self.tree is None:
            return None
        # Note that we return the "lemma category", which is suitable for
        # topic indexing and similar applications. Unknown words and entity
        # names have the category 'entity' in this case, and person names
        # have one of 'person_kk'/'person_kvk'/'person_hk'.
        # pylint: disable=not-an-iterable
        return [(d.lemma, d.lemma_cat) for d in self.terminal_nodes]

    @property
    def ifd_tags(self) -> Optional[List[str]]:
        """ Return a list of Icelandic Frequency Dictionary (IFD) tags for
            the terminals/tokens in this sentence. """
        if self.tree is None:
            return None
        # Flatten the ifd_tags lists for the individual nodes
        # (nonterminal nodes return an empty list in the ifd_tags property)
        return [ifd_tag for d in self.tree.descendants for ifd_tag in d.ifd_tags]

    def dump(self, greynir_cls: GreynirType) -> Dict[str, Any]:
        """ Dump internal data of the class instance for serialization.
            Useful for storing parsed data in a database.
            Note: Normally, sentences are dumped using Greynir.dumps_single(). """
        return {
            "tokens": [greynir_cls._dump_token(t) for t in self._s],
            "tree": None if self.tree is None else self.tree._head,
        }

    def dumps(self, greynir_cls: GreynirType, **kwargs: Any) -> str:
        """ Dump internal data of the class instance as a json string.
            Useful for storing parsed data in a database.
            Note: Normally, sentences are dumped using Greynir.dumps_single(). """
        if "ensure_ascii" not in kwargs:
            # Unless explicitly stated, we are OK with UTF-8 in the generated JSON
            return json.dumps(self.dump(greynir_cls), ensure_ascii=False, **kwargs)
        return json.dumps(self.dump(greynir_cls), **kwargs)

    @classmethod
    def load(
        cls,
        greynir_cls: GreynirType,
        tokens: List[Sequence[Any]],
        tree: Optional[CanonicalTokenDict],
    ) -> "_Sentence":
        """ Load previously dumped data.
            Useful for retrieving parsed data from a database.
            Note: Normally, sentences are loaded using Greynir.loads_single(). """
        instance = cls.__new__(cls)
        instance.__dict__ = {
            "_s": [greynir_cls._load_token(*t) for t in tokens],
            "len": len(tokens),
            "_simplified_tree": None if tree is None else SimpleTree([[tree]]),
            "_terminals": None,
            "_job": None,
            "_err_index": None,
            "_error": None,
            "_score": None,
        }
        return instance

    @classmethod
    def loads(
        cls, greynir_cls: GreynirType, json_str: str, **kwargs: Any
    ) -> "_Sentence":
        """ Load a previously dumped JSON string.
            Useful for retrieving parsed data from a database.
            Note: Normally, sentences are loaded using Greynir.loads_single(). """
        data = json.loads(json_str, **kwargs)
        return cls.load(greynir_cls, **data)

    def __str__(self) -> str:
        """ Return a text representation of a sentence """
        return self.text


# Create a public alias for the _Sentence class
# Note that putting a type annotation here (i.e. Sentence: Type[_Sentence] = _Sentence)
# is not a good idea, cf. https://mypy.readthedocs.io/en/latest/common_issues.html#variables-vs-type-aliases
Sentence = _Sentence


class _NounPhrase(_Sentence):

    """ A specialization for parsed noun phrases,
        providing easy access to inflectional forms """

    _nom: GetterFunc = operator.attrgetter("nominative_np")
    _acc: GetterFunc = operator.attrgetter("accusative_np")
    _dat: GetterFunc = operator.attrgetter("dative_np")
    _gen: GetterFunc = operator.attrgetter("genitive_np")
    _ind: GetterFunc = operator.attrgetter("indefinite_np")
    _can: GetterFunc = operator.attrgetter("canonical_np")

    def _get(self, getter: GetterFunc) -> Optional[str]:
        if self.tree is None:
            return None
        return correct_spaces(getter(self.tree))

    @cached_property
    def nominative(self) -> Optional[str]:
        """ Return nominative form (nefnifall) """
        return self._get(_NounPhrase._nom)

    @cached_property
    def indefinite(self) -> Optional[str]:
        """ Return indefinite form (nefnifall án greinis) """
        return self._get(_NounPhrase._ind)

    @cached_property
    def canonical(self) -> Optional[str]:
        """ Return canonical form (nefnifall eintölu án greinis) """
        return self._get(_NounPhrase._can)

    @cached_property
    def accusative(self) -> Optional[str]:
        """ Return accusative form (þolfall) """
        return self._get(_NounPhrase._acc)

    @cached_property
    def dative(self) -> Optional[str]:
        """ Return dative form (þágufall) """
        return self._get(_NounPhrase._dat)

    @cached_property
    def genitive(self) -> Optional[str]:
        """ Return genitive form (eignarfall) """
        return self._get(_NounPhrase._gen)


class _Paragraph:

    """ Encapsulates a paragraph that contains sentences """

    def __init__(self, job: "_Job", p: Iterable[SentenceTuple]) -> None:
        self._job = job
        self._p = p

    def sentences(self) -> Iterable[_Sentence]:
        """ Yield the sentences within the paragraph, nicely wrapped """
        # self._p is a generator that yields (ix, toklist) tuples,
        # where ix is a starting index of the sentence within the
        # token stream, and toklist is the list of tokens in the sentence,
        # not including S_BEGIN and S_END tokens.
        for _, sent in self._p:
            yield self._job._create_sentence(sent)

    def __iter__(self) -> Iterator[_Sentence]:
        """ Allow easy iteration of sentences within this paragraph """
        return iter(self.sentences())


# Create a public alias for the _Paragraph class
Paragraph = _Paragraph


class _Job:

    """ A parsing job object, allowing incremental parsing of text
        by paragraph and/or sentence.
    """

    def __init__(
        self,
        greynir: "Greynir",
        tokens: Iterable[Tok],
        *,
        parse: bool = False,
        root: Optional[str] = None,
        progress_func: ProgressFunc = None,
        max_sent_tokens: int = DEFAULT_MAX_SENT_TOKENS,
    ) -> None:
        self._r = greynir
        self._parser = self._r.parser
        self._reducer = self._r.reducer
        self._tokens = tokens
        self._parse_time = 0.0
        self._reduce_time = 0.0
        self._parse = parse
        # Pre-counted number of sentences (used for progress monitoring)
        # This count includes one initial step, indicating completion of
        # tokenization, so it is the number of sentences + 1
        self._cnt_sent = 0
        # Accumulated number of sentences generated so far
        self._num_sent = 0
        self._num_parsed = 0
        self._num_tokens = 0
        self._num_combinations = 0
        self._total_ambig = 0.0
        self._total_tokens = 0
        # The grammar root nonterminal to be used
        # for parsing within this job
        self._root = root
        # A progress function to call during processing
        self._progress_func = progress_func
        # The maximum length, in tokens, of a sentence that we will attempt to parse
        self._max_sent_tokens = max_sent_tokens

    def _add_sentence(
        self, s: TokenList, num: int, parse_time: float, reduce_time: float
    ) -> None:
        """ Add a processed sentence to the statistics """
        slen = len(s)
        self._num_sent += 1
        self._num_tokens += slen
        if num > 0:
            # The sentence was parsed successfully
            self._num_parsed += 1
            self._num_combinations += num
            ambig_factor = num ** (1 / slen)
            self._total_ambig += ambig_factor * slen
            self._total_tokens += slen
        # Accumulate the total time spent on parsing and reduction
        self._parse_time += parse_time
        # Accumulate the time thereof spent on reduction
        self._reduce_time += reduce_time
        # Call the progress function, if any
        if self._progress_func is not None:
            assert self._cnt_sent > 0
            assert (self._num_sent + 1) <= self._cnt_sent
            # We add one to the _num_sent counter because we
            # have an additional first step that we call after tokenization
            self._progress_func((self._num_sent + 1) / self._cnt_sent)

    def _create_sentence(self, s: TokenList) -> _Sentence:
        """ Create a fresh _Sentence object """
        return self._r.create_sentence(self, s)

    @property
    def parse_immediately(self) -> bool:
        """ Return True if sentences in the job should be parsed immediately """
        return self._parse

    def paragraphs(self) -> Iterable[_Paragraph]:
        """ Yield the paragraphs from the token stream """
        if self._progress_func is not None:
            # We have a progress function, so we must pre-count
            # the sentences to be processed. This means that all input
            # data must be converted to lists, exhausting generators.
            # Note that the paragraph splitting phase applies the
            # token-level corrections, so they are not really
            # counted in the progress.
            pg_list = [
                [(ix, list(sent)) for ix, sent in p] for p in paragraphs(self._tokens)
            ]
            self._cnt_sent = sum(len(p) for p in pg_list) + 1
            # Make an "first step" initial call to the progress function
            # with a progress of 1, after we've tokenized the input
            self._progress_func(1 / self._cnt_sent)
            plist = iter(pg_list)
        else:
            # No progress function: use generators throughout
            plist = paragraphs(self._tokens)
        for p in plist:
            yield _Paragraph(self, p)

    def sentences(self) -> Iterable[_Sentence]:
        """ Yield the sentences from the token stream """
        for p in self.paragraphs():
            yield from p.sentences()

    def parse(self, tokens: TokenList) -> Tuple[Any, int, int]:
        """ Parse the token sequence, returning a parse tree,
            the number of trees in the parse forest, and the
            score of the best tree """
        num = 0
        score = 0
        forest = None
        t0 = t1 = time.time()
        try:
            if self._max_sent_tokens and len(tokens) > self._max_sent_tokens:
                # Sentence is above the maximum length: don't attempt to parse it
                raise ParseError(
                    "Sentence is longer than {0} tokens".format(self._max_sent_tokens),
                    token_index=self._max_sent_tokens,
                )
            if not self.parse_foreign_sentences and tokens_are_foreign(
                tokens, min_icelandic_ratio=ICELANDIC_RATIO
            ):
                # Sentence is foreign: don't attempt to parse it
                raise ParseError("Sentence is probably not in Icelandic", token_index=0)
            forest = self.parser.go(tokens, root=self._root)
            t1 = time.time()
            if forest is not None:
                num = Fast_Parser.num_combinations(forest)
                if num > 1:
                    # Reduce the parse forest to a single
                    # "best" (highest-scoring) parse tree
                    forest, score = self.reducer.go_with_score(forest)
            return forest, num, score
        finally:
            # Accumulate statistics in the job object
            now = time.time()
            self._add_sentence(tokens, num, parse_time=now - t0, reduce_time=now - t1)

    def __iter__(self) -> Iterator[_Sentence]:
        """ Allow easy iteration of sentences within this job """
        return iter(self.sentences())

    @property
    def parser(self) -> Fast_Parser:
        """ The job's associated parser object """
        return self._parser

    @property
    def reducer(self) -> Reducer:
        """ The job's associated reducer object """
        return self._reducer

    @property
    def num_tokens(self) -> int:
        """ Total number of tokens in sentences submitted to this job """
        return self._num_tokens

    @property
    def num_sentences(self) -> int:
        """ Total number of sentences submitted to this job """
        return self._num_sent

    @property
    def num_parsed(self) -> int:
        """ Total number of sentences successfully parsed within this job """
        return self._num_parsed

    @property
    def num_combinations(self) -> int:
        """ Sum of the total number of parse tree combinations
            for sentences within this job """
        return self._num_combinations

    @property
    def ambiguity(self) -> float:
        """ The weighted average total ambiguity of parsed sentences
            within this job """
        return (
            (self._total_ambig / self._total_tokens) if self._total_tokens > 0 else 1.0
        )

    @property
    def parse_time(self) -> float:
        """ Total time spent on parsing (including reduction) during this job,
            in seconds """
        return self._parse_time

    @property
    def reduce_time(self) -> float:
        """ Total time spent on tree reduction during this job, in seconds """
        return self._reduce_time

    @property
    def parse_foreign_sentences(self) -> bool:
        """ Return True if foreign-looking sentences should be parsed """
        return self._r.parse_foreign_sentences


# Create a public alias for the _Job class
Job = _Job


class _Job_NP(_Job):

    """ Specialized _Job class that creates _NounPhrase objects
        instead of _Sentence objects """

    def __init__(
        self,
        greynir: "Greynir",
        tokens: Iterable[Tok],
        *,
        force_number: Optional[str] = None,
    ) -> None:
        # Parse the tokens with 'Nl' (noun phrase) as the root nonterminal
        # instead of the usual default 'S0' (sentence) root
        root = "Nl"
        if force_number:
            if force_number in ("et", "singular"):
                # We only want singular parses
                root = "Nl_et"
            elif force_number in ("ft", "plural"):
                # We only want plural parses
                root = "Nl_ft"
            else:
                raise ValueError("Unexpected value of force_number")
        super().__init__(greynir, tokens, parse=True, root=root)

    def _create_sentence(self, s: TokenList) -> _NounPhrase:
        """ Create a fresh _NounPhrase object """
        return _NounPhrase(self, s)


class Greynir:

    """ Utility class to tokenize and parse text, organized
        as a sequence of sentences or alternatively as paragraphs
        of sentences. Typical usage:

        g = Greynir()
        job = g.submit(my_text)
        # Iterate through sentences and parse each one:
        for sent in job:
            if sent.parse():
                # sentence parsed successfully
                # do something with sent.tree
                print(sent.tree)
            else:
                # an error occurred in the parse
                # the error token index is at sent.err_index
                pass
        # Alternatively, split into paragraphs first:
        job = g.submit(my_text)
        for p in job.paragraphs(): # Yields paragraphs
            for sent in p.sentences(): # Yields sentences
                if sent.parse():
                    # sentence parsed successfully
                    # do something with sent.tree
                    print(sent.tree)
                else:
                    # an error occurred in the parse
                    # the error token index is at sent.err_index
                    pass
        # After parsing all sentences in a job, the following
        # statistics are available:
        num_sentences = job.num_sentences   # Total number of sentences
        num_parsed = job.num_parsed         # Thereof successfully parsed
        ambiguity = job.ambiguity           # Average ambiguity factor
        parse_time = job.parse_time         # Elapsed time since job was created

    """

    _parser: Optional[Fast_Parser] = None
    _reducer: Optional[Reducer] = None
    _lock = Lock()

    def __init__(self, **options: Any) -> None:
        """ Tokenization options can be passed as keyword arguments to the
            Greynir constructor """
        # Set parse_foreign_sentences to True to attempt to parse
        # all sentences, even if probably foreign
        self._parse_foreign_sentences: bool = options.pop(
            "parse_foreign_sentences", False
        )
        self._options = options

    @property
    def parse_foreign_sentences(self) -> bool:
        """ Return True if the parser should attempt to parse sentences
            that look to be foreign, i.e. not in Icelandic """
        return self._parse_foreign_sentences

    @classmethod
    def _dump_token(cls, tok: Tok) -> Tuple[Any, ...]:
        """ Allow derived classes to override how tokens are dumped """
        # Returns (kind, txt, val) - corresponding to
        # the expected signature of _load_token()
        return (tok.kind, tok.txt, tok.val)

    @classmethod
    def _load_token(cls, *args: Any) -> Tok:
        """ Load token from serialized data """
        return Tok(*load_token(*args))

    def dumps_single(self, sent: _Sentence, **kwargs: Any) -> str:
        """ Return a _Sentence object in a JSON-formatted string,
            which can be loaded again using loads_single() """
        return sent.dumps(self.__class__, **kwargs)

    def loads_single(self, json_str: str, **kwargs: Any) -> _Sentence:
        """ Load previously dumped JSON description of a single sentence.
            Useful for retrieving parsed data from a database. """
        return _Sentence.loads(self.__class__, json_str, **kwargs)

    def tokenize(self, text: StringIterable) -> Iterable[Tok]:
        """ Call the tokenizer (overridable in derived classes) """
        return bin_tokenize(text, **self._options)

    def create_sentence(self, job: _Job, s: TokenList) -> _Sentence:
        """ Override this in derived classes to modify how sentences
            are created or postprocessed """
        return _Sentence(job, s)

    @property
    def parser(self) -> Fast_Parser:
        """ Return the parser instance to be used """
        with self._lock:
            if Greynir._parser is None:
                # Initialize a singleton instance of the parser and the reducer.
                # Both classes are re-entrant and thread safe.
                Greynir._parser = Fast_Parser()
                Greynir._reducer = Reducer(Greynir._parser.grammar)
            return Greynir._parser

    @property
    def reducer(self) -> Reducer:
        """ Return the reducer instance to be used """
        # Should always retrieve the parser attribute first
        assert Greynir._reducer is not None
        return Greynir._reducer

    def submit(
        self,
        text: StringIterable,
        parse: bool = False,
        *,
        split_paragraphs: bool = False,
        progress_func: ProgressFunc = None,
        max_sent_tokens: int = DEFAULT_MAX_SENT_TOKENS,
    ) -> _Job:
        """ Submit a text to the tokenizer and parser, yielding a job object.
            The paragraphs and sentences of the text can then be iterated
            through via the job object. If parse is set to True, the
            sentences are automatically parsed before being returned.
            Otherwise, they need to be explicitly parsed by calling
            sent.parse(). This is a more incremental, asynchronous
            approach than Greynir.parse().
            
            If progress_func is given, it will be called during processing
            with a single float parameter between 0.0..1.0 indicating the
            ratio of progress so far with the parsing job. """

        if split_paragraphs:
            # Original text consists of paragraphs separated by newlines:
            # insert paragraph separators before tokenization
            assert isinstance(text, str)
            text = mark_paragraphs(text)
        tokens = self.tokenize(text)
        return _Job(
            self,
            tokens,
            parse=parse,
            progress_func=progress_func,
            max_sent_tokens=max_sent_tokens,
        )

    def parse(
        self,
        text: str,
        *,
        progress_func: ProgressFunc = None,
        max_sent_tokens: int = DEFAULT_MAX_SENT_TOKENS,
    ) -> ParseResult:
        """ Convenience function to parse text synchronously and return
            a summary of all contained sentences. The progress_func parameter
            works as described for Greynir.submit(). """
        tokens = self.tokenize(text)
        job = _Job(
            self,
            tokens,
            parse=True,
            progress_func=progress_func,
            max_sent_tokens=max_sent_tokens,
        )
        # Iterating through the sentences in the job causes
        # them to be parsed and their statistics collected
        sentences = [sent for sent in job]
        return ParseResult(
            sentences=sentences,
            num_sentences=job.num_sentences,
            num_parsed=job.num_parsed,
            num_tokens=job.num_tokens,
            ambiguity=job.ambiguity,
            parse_time=job.parse_time,
            reduce_time=job.reduce_time,
        )

    def parse_single(
        self, sentence: str, *, max_sent_tokens: int = DEFAULT_MAX_SENT_TOKENS
    ) -> Optional[_Sentence]:
        """ Convenience function to parse a single sentence only """
        tokens = self.tokenize(sentence)
        job = _Job(self, tokens, parse=True, max_sent_tokens=max_sent_tokens)
        # Returns None if no sentence could be extracted from the text
        try:
            return next(iter(job))
        except StopIteration:
            return None

    def parse_tokens(
        self, tokens: Iterable[Tok], *, max_sent_tokens: int = DEFAULT_MAX_SENT_TOKENS
    ) -> Optional[_Sentence]:
        """ Convenience function to parse a single sentence from tokens """
        job = _Job(self, tokens, parse=True, max_sent_tokens=max_sent_tokens)
        # Returns None if no sentence could be extracted from the text
        try:
            return next(iter(job))
        except StopIteration:
            return None

    def parse_noun_phrase(
        self, noun_phrase: str, *, force_number: Optional[str] = None
    ) -> Optional[_NounPhrase]:
        """ Utility function to parse a noun phrase. Note that in most
            cases it is more convenient to use the NounPhrase class
            for this purpose. """
        # When tokenizing a noun phrase, don't assume that it starts a sentence
        tokens = self.tokenize(noun_phrase)
        # Use a _Job_NP to generate _NounPhrase objects instead of _Sentence objects
        job = _Job_NP(self, tokens, force_number=force_number)
        # Returns None if no noun phrase could be extracted from the text
        try:
            return cast(_NounPhrase, next(iter(job)))
        except StopIteration:
            return None

    def lemmatize(
        self,
        txt: str,
        *,
        all_lemmas: bool = False,
        sortkey: Optional[Callable[[LemmaTuple], Comparable]] = None,
    ) -> Union[Iterator[LemmaTuple], Iterator[List[LemmaTuple]]]:
        """ Utiility function to (simplistically) lemmatize all words in
            a given string without parsing. Returns a generator of
            (lemma, word category) tuples, one for each text token
            in the input (text tokens being words, person names and entity names).
            Punctuation, dates, numbers, e-mail addresses and other token types
            are skipped and not included in the output.
            If all_lemmas is True, the function returns a list of tuples
            with all possible lemmas for each text token.
            If all_lemmas is True and a sortkey is given, the returned
            list is sorted using that function as a sort key, cf. the
            Python built-in list.sort() function. """
        return simple_lemmatize(txt, all_lemmas=all_lemmas, sortkey=sortkey)

    @classmethod
    def cleanup(cls) -> None:
        """ Discard memory resources held by the Greynir class object """
        cls._reducer = None
        if cls._parser is not None:
            Fast_Parser.discard_grammar()
            cls._parser.cleanup()
            cls._parser = None


# Allow old class name for compatibility
Reynir = Greynir



"""
    Greynir: Natural language processing for Icelandic

    Basic classes module

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


    This module contains basic functions that are used by the settings
    module and other modules. These functions have been extracted from the
    settings module to avoid circular imports or module references.

"""

from typing import (
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
)

import os
import locale

from contextlib import contextmanager
from pkg_resources import resource_stream


# The locale used by default in the changedlocale function
_DEFAULT_LOCALE = ("IS_is", "UTF-8")

# A set of all valid verb argument cases
ALL_CASES = frozenset(("nf", "þf", "þgf", "ef"))
ALL_GENDERS = frozenset(("kk", "kvk", "hk"))
ALL_NUMBERS = frozenset(("et", "ft"))
SUBCLAUSES = frozenset(("nh", "mnh", "falls"))
REFLPRN = {"sig": "sig_hk_et_þf", "sér": "sig_hk_et_þgf", "sín": "sig_hk_et_ef"}
REFLPRN_SET = frozenset(REFLPRN.keys())

# BÍN compressed file format version (used in tools/binpack.py and bincompress.py)
BIN_COMPRESSOR_VERSION = b"Greynir 02.00.00"
assert len(BIN_COMPRESSOR_VERSION) == 16
BIN_COMPRESSED_FILE = "ord.compressed"


@contextmanager
def changedlocale(
    new_locale: Optional[str] = None, category: str = "LC_COLLATE"
) -> Iterator[Callable[[str], str]]:
    """ Change locale for collation temporarily within a context (with-statement) """
    # The newone locale parameter should be a tuple: ('is_IS', 'UTF-8')
    # The category should be a string such as 'LC_TIME', 'LC_NUMERIC' etc.
    cat = getattr(locale, category)
    old_locale = locale.getlocale(cat)
    try:
        locale.setlocale(cat, new_locale or _DEFAULT_LOCALE)
        yield locale.strxfrm  # Function to transform string for sorting
    finally:
        locale.setlocale(cat, old_locale)


def sort_strings(strings: Iterable[str], loc: Optional[str] = None) -> List[str]:
    """ Sort a list of strings using the specified locale's collation order """
    # Change locale temporarily for the sort
    with changedlocale(loc) as strxfrm:
        return sorted(strings, key=strxfrm)


class ConfigError(Exception):

    """ Exception class for configuration errors """

    def __init__(self, s: str) -> None:
        super().__init__(s)
        self.fname: Optional[str] = None
        self.line = 0

    def set_pos(self, fname: str, line: int) -> None:
        """ Set file name and line information, if not already set """
        if not self.fname:
            self.fname = fname
            self.line = line

    def __str__(self) -> str:
        """ Return a string representation of this exception """
        s = Exception.__str__(self)
        if not self.fname:
            return s
        return "File {0}, line {1}: {2}".format(self.fname, self.line, s)


class LineReader:

    """ Read lines from a text file, recognizing $include directives """

    def __init__(
        self,
        fname: str,
        *,
        package_name: Optional[str] = None,
        outer_fname: Optional[str] = None,
        outer_line: int = 0
    ) -> None:
        self._fname = fname
        self._package_name = package_name
        self._line = 0
        self._inner_rdr: Optional[LineReader] = None
        self._outer_fname = outer_fname
        self._outer_line = outer_line

    def fname(self) -> str:
        """ The name of the file being read """
        return self._fname if self._inner_rdr is None else self._inner_rdr.fname()

    def line(self) -> int:
        """ The number of the current line within the file """
        return self._line if self._inner_rdr is None else self._inner_rdr.line()

    def lines(self) -> Iterator[str]:
        """ Generator yielding lines from a text file """
        self._line = 0
        try:
            if self._package_name:
                stream = resource_stream(self._package_name, self._fname)
            else:
                stream = open(self._fname, "rb")
            with stream as inp:
                # Read config file line-by-line from the package resources
                accumulator = ""
                for b in inp:
                    # We get byte strings; convert from utf-8 to Python strings
                    s = b.decode("utf-8")
                    self._line += 1
                    if s.rstrip().endswith("\\"):
                        # Backslash at end of line: continuation in next line
                        accumulator += s.strip()[:-1]
                        continue
                    if accumulator:
                        # Add accumulated text from preceding
                        # backslash-terminated lines, but drop leading whitespace
                        s = accumulator + s.lstrip()
                        accumulator = ""
                    # Check for include directive: $include filename.txt
                    if s.startswith("$") and s.lower().startswith("$include "):
                        iname = s.split(maxsplit=1)[1].strip()
                        # Do some path magic to allow the included path
                        # to be relative to the current file path, or a
                        # fresh (absolute) path by itself
                        head, _ = os.path.split(self._fname)
                        iname = os.path.join(head, iname)
                        rdr = self._inner_rdr = LineReader(
                            iname,
                            package_name=self._package_name,
                            outer_fname=self._fname,
                            outer_line=self._line,
                        )
                        yield from rdr.lines()
                        self._inner_rdr = None
                    else:
                        yield s
                if accumulator:
                    # Catch corner case where last line of file ends with a backslash
                    yield accumulator
        except (IOError, OSError):
            if self._outer_fname:
                # This is an include file within an outer config file
                c = ConfigError(
                    "Error while opening or reading include file '{0}'".format(
                        self._fname
                    )
                )
                c.set_pos(self._outer_fname, self._outer_line)
            else:
                # This is an outermost config file
                c = ConfigError(
                    "Error while opening or reading config file '{0}'".format(
                        self._fname
                    )
                )
            raise c



"""

    Greynir: Natural language processing for Icelandic

    NounPhrase class implementation

    Copyright (C) 2021 Miðeind ehf.
    Original author: Vilhjálmur Þorsteinsson

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module implements the NounPhrase class, a handy container
    for noun phrases (nafnliður) allowing them to be easily inflected
    and formatted.

"""

from typing import Optional, Mapping, Callable

import operator

from .reynir import Greynir, _NounPhrase, SimpleTree


# Format specifiers and how they relate to properties
# of the contained NounPhrase object
_FMT: Mapping[str, Callable[["_NounPhrase"], str]] = {
    # Icelandic format specifiers
    "nf": operator.attrgetter('nominative'),
    "þf": operator.attrgetter('accusative'),
    "þgf": operator.attrgetter('dative'),
    "ef": operator.attrgetter('genitive'),
    "ángr": operator.attrgetter('indefinite'),
    "stofn": operator.attrgetter('canonical'),
    # English/international format specifiers
    "nom": operator.attrgetter('nominative'),
    "acc": operator.attrgetter('accusative'),
    "dat": operator.attrgetter('dative'),
    "gen": operator.attrgetter('genitive'),
    "ind": operator.attrgetter('indefinite'),
    "can": operator.attrgetter('canonical'),
}


class NounPhrase:

    """ A handy container for a noun phrase (nafnliður),
        allowing it to be easily inflected and formatted """

    # Singleton parser instance
    _greynir: Optional[Greynir] = None

    def __init__(self, np_string: str, *, force_number: Optional[str] = None) -> None:
        """ Initialize a NounPhrase from a text string.
            If force_number is set to "et" or "singular", we only
            consider singular interpretations of the string.
            If force_number is set to "ft" or "plural", we only
            consider plural interpretations of the string. """
        self._np_string = np_string or ""
        self._number: Optional[str] = None
        self._person: Optional[str] = None
        self._case: Optional[str] = None
        self._gender: Optional[str] = None
        self._np: Optional[_NounPhrase] = None
        if self._np_string:
            if self._greynir is None:
                # Initialize our parser singleton
                # When parsing noun phrases, we don't assume that they
                # start a sentence - so we don't attempt to interpret the
                # first word as a lowercase word, as we would otherwise
                self.__class__._greynir = Greynir(no_sentence_start=True)
            # Parse the noun phrase string into a _NounPhrase object
            assert self._greynir is not None
            self._np = self._greynir.parse_noun_phrase(
                self._np_string, force_number=force_number
            )
            if self._np is not None and self._np.deep_tree is not None:
                # Access the first child of the root 'Nl' nonterminal
                # of the deep parse tree
                nt = next(self._np.deep_tree.enum_child_nodes()).nonterminal.name
                # Sanity check
                assert nt.startswith("Nl_") or nt.startswith("NlEind_")
                # Extract the variants of the nonterminal
                variants = set(nt.split("_")[1:])
                self._number = (variants & {"et", "ft"}).pop()
                self._person = (variants & {"p1", "p2", "p3"}).pop()
                self._case = (variants & {"nf", "þf", "þgf", "ef"}).pop()
                self._gender = (variants & {"kk", "kvk", "hk"}).pop()

    def __str__(self) -> str:
        """ Return the contained string as-is """
        return self._np_string

    def __repr__(self) -> str:
        return "<reynir.NounPhrase('{0}'), {1}>".format(
            self._np_string,
            "parsed" if self.parsed else "not parsed"
        )

    def __len__(self) -> int:
        """ Provide len() for convenience """
        return self._np_string.__len__()

    def __format__(self, format_spec: str) -> str:
        """ Return the contained string after inflecting it according
            to the format specification, if given """
        # Examples:
        # >>> np = NounPhrase('skjótti hesturinn')
        # >>> f"Hér er {np:nf}"
        # 'Hér er skjótti hesturinn'
        # >>> f"Um {np:þf}"
        # 'Um skjótta hestinn'
        # >>> f"Frá {np:þgf}"
        # 'Frá skjótta hestinum'
        # >>> f"Til {np:ef}"
        # 'Til skjótta hestsins'
        # >>> f"Hér er {np:ángr}"
        # 'Hér er skjóttur hestur'
        # np = NounPhrase("þrír skjóttir hestar")
        # >>> f"Umræðuefnið er {np:stofn}"
        # 'Umræðuefnið er skjóttur hestur'
        if not format_spec or not self.parsed:
            return self._np_string
        # Find the attrgetter (property access function)
        # corresponding to the format spec
        fmt = _FMT.get(format_spec)
        if fmt is None:
            # We don't recognize this format specifier
            raise ValueError(
                "Invalid format specifier for NounPhrase: '{0}'".format(format_spec)
            )
        # Extract the requested property and return it
        assert self._np is not None
        return fmt(self._np)

    @property
    def parsed(self) -> bool:
        """ Return True if the noun phrase was successfully parsed """
        return self._np is not None and self._np.tree is not None

    @property
    def tree(self) -> Optional[SimpleTree]:
        """ Return the SimpleTree object corresponding to the noun phrase """
        return None if self._np is None else self._np.tree

    @property
    def case(self) -> Optional[str]:
        """ Return the case of the noun phrase, as originally parsed """
        return self._case

    @property
    def number(self) -> Optional[str]:
        """ Return the number (singular='et'/plural='ft') of the noun phrase,
            as originally parsed """
        return self._number

    @property
    def person(self) -> Optional[str]:
        """ Return the person ('p1', 'p2', 'p3') of the noun phrase,
            as originally parsed """
        return self._person

    @property
    def gender(self) -> Optional[str]:
        """ Return the gender (masculine='kk', feminine='kvk', neutral='hk')
            of the noun phrase, as originally parsed """
        return self._gender

    @property
    def nominative(self) -> Optional[str]:
        """ Return nominative form (nefnifall) """
        return None if self._np is None else self._np.nominative

    @property
    def indefinite(self) -> Optional[str]:
        """ Return indefinite form (nefnifall án greinis) """
        return None if self._np is None else self._np.indefinite

    @property
    def canonical(self) -> Optional[str]:
        """ Return canonical form (nefnifall eintölu án greinis) """
        return None if self._np is None else self._np.canonical

    @property
    def accusative(self) -> Optional[str]:
        """ Return accusative form (þolfall) """
        return None if self._np is None else self._np.accusative

    @property
    def dative(self) -> Optional[str]:
        """ Return dative form (þágufall) """
        return None if self._np is None else self._np.dative

    @property
    def genitive(self) -> Optional[str]:
        """ Return genitive form (eignarfall) """
        return None if self._np is None else self._np.genitive




"""
    Greynir: Natural language processing for Icelandic

    Parser base module

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module defines a base parser class. The base is used in
    BIN_Parser (see binparser.py) which is again the base of the
    C++ Earley parser Fast_Parser (see fastparser.py)

"""

from typing import Dict, List, Iterator, Optional

from .grammar import Grammar, GrammarItem, Terminal, Nonterminal, Production


class _PackedProduction:

    """ A container for a packed production, i.e. a grammar Production
        where the component terminals and nonterminals have been packed
        into a list of integer indices """

    def __init__(self, priority: int, production: Production) -> None:
        # Store the relative priority of this production within its nonterminal
        self._priority = priority
        # Keep a reference to the original production
        self._production = production
        # Store the packed list of indices
        self._ix_list = production.prod
        # Cache the length
        self._len = len(self._ix_list)

    @property
    def production(self) -> Production:
        return self._production

    @property
    def priority(self) -> int:
        return self._priority

    def __getitem__(self, index: int) -> int:
        return self._ix_list[index] if 0 <= index < self._len else 0

    def __len__(self) -> int:
        return self._len

    def __iter__(self) -> Iterator[int]:
        return iter(self._ix_list)


class Base_Parser:

    """ Parses a sequence of tokens according to a given grammar and
        a root nonterminal within that grammar, returning a forest of
        possible parses. The parses uses an optimized Earley algorithm.
    """

    def __init__(self) -> None:
        self._root: Optional[int] = None
        self._nt_dict: Dict[int, Optional[List[_PackedProduction]]] = {}
        self._nonterminals: Dict[int, Nonterminal] = {}
        self._terminals: Dict[int, Terminal] = {}

    def init_from_grammar(self, g: Grammar) -> None:
        """ Initialize the parser with the given grammar """
        nt_d = g.nt_dict
        r = g.root
        assert nt_d is not None
        assert r is not None
        assert r in nt_d
        # Convert the grammar to integer index representation for speed
        self._root = r.index
        # Make new grammar dictionary, keyed by nonterminal index and
        # containing packed productions with integer indices
        self._nt_dict = {}
        for nt, plist in nt_d.items():
            self._nt_dict[nt.index] = (
                None
                if plist is None
                else [_PackedProduction(prio, p) for prio, p in plist]
            )
        self._nonterminals = g.nonterminals_by_ix
        self._terminals = g.terminals_by_ix

    @classmethod
    def for_grammar(cls, g: Grammar) -> "Base_Parser":
        """ Create a parser for the Grammar in g """
        p = cls()
        p.init_from_grammar(g)
        return p

    def _lookup(self, ix: int) -> GrammarItem:
        """ Convert a production item from an index to an object reference """
        # Terminals have positive indices
        # Nonterminals have negative indices
        # A zero index is not allowed
        assert ix != 0
        return self._nonterminals[ix] if ix < 0 else self._terminals[ix]
